
from beanie import Document
from beanie import init_beanie

from motor.motor_asyncio import AsyncIOMotorClient

from .agent import AgentResponse
from .chat import Chat, ChatToolState
from .message import Message
from .storage import Storage, StorageItem
from .tool import Tool, ToolAnswer


async def setup_database(uri: str, name: str):
    client = AsyncIOMotorClient(uri)[name]
    await init_beanie(client, document_models=Document.__subclasses__())




import enum


class ChatStatus(enum.Enum):
    created = "created"
    error = "error"
    idle = "idle"
    tools = "tools"
    generating = "generating"
    finished = "finished"


class ChatToolStateStatus(enum.Enum):
    in_progress = "in_progress"
    completed = "completed"
    error = "error"




from beanie import Document

from .meta import TimestampMixin


class Tool(TimestampMixin, Document):
    tool_id: str
    name: str
    arguments: str


class ToolAnswer(TimestampMixin, Document):
    tool_id: str
    name: str
    text: str




from . import create
from . import start
from . import tool

import click
import os

from AgentService.utils.checkers import is_project
from AgentService.utils.logger import setup_logger
from AgentService.config import Config


__execute = click.CommandCollection(
    sources=[
        create.group
    ],
    help='Use "agent <command> -h/--help" to see more info about a command',
)

__execute_project = click.CommandCollection(
    sources=[
        tool.group,
        start.group
    ],
    help='Use "python3.11 manage.py <command> -h/--help" to see more info about a command',
)


def execute():
    __execute()


def execute_project():
    current_path = os.getcwd()
    if not is_project(current_path):
        print(f'Error: no AgentService project found in {current_path}')
        return

    config = Config(current_path)
    setup_logger(config.log_path, config.log_level)

    __execute_project()




import asyncio
import enum
import functools
import typing
import aiohttp

from loguru import logger

from .event import ToolEvent

from AgentService.types import (
    AgentResponse, Tool,
    ToolAnswer, ChatToolState
)

from AgentService.enums.tool import ToolResponse
from AgentService.enums.agent import AgentResponseType

from AgentService.app.routes.chat.models import SendMessageRequest


class AgentConnector:
    def __init__(
        self,
        endpoint: str,
        key: str = None
    ):

        self.log = logger.bind(classname=self.__class__.__name__)

        self.endpoint = endpoint[:-1] if endpoint.endswith("/") else endpoint
        self.key = key

        self.callbacks = {}

    def bind_tool_output(self, tool_name: str, function: typing.Callable):
        self.callbacks.update({tool_name: functools.partial(self.answer_tool, function=function)})
        self.log.debug(f"Binded callback -> {tool_name} -> {function}")

    async def add_to_storage(self, key: str, data: list[dict] | dict):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/storage/add",
                json={
                    "key": key,
                    "data": data
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def remove_from_storage(self, key: str, data: list[dict] | dict):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/storage/remove",
                json={
                    "key": key,
                    "data": data
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def update_storage(self, key: str, data: dict, new_data: dict):
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/storage/update",
                json={
                    "key": key,
                    "data": data,
                    "new_data": new_data
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def get_from_storage(self, key: str) -> list[dict]:
        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=f"{self.endpoint}/storage",
                json={
                    "key": key
                },
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")
                return await resp.json()

    async def answer_tool(self, event: ToolEvent, function: typing.Callable) -> ToolAnswer:
        try:
            response = await function(data=event.tool.arguments, chat_id=event.chat_id)

            if response in [None, ""]:
                response = "function returned nothing"

            elif isinstance(response, enum.Enum):
                response = response.value

            elif not isinstance(response, str):
                response = str(response)

            return ToolAnswer(
                tool_id=event.tool.tool_id,
                name=event.tool.name,
                text=response
            )

        except Exception as err:
            self.log.exception(err)

            return ToolAnswer(
                tool_id=event.tool.tool_id,
                name=event.tool.name,
                text="function returned nothing"
            )

    async def dispatch_event(self, event: ToolEvent):
        callback = self.callbacks.get(event.tool.name)

        if not callback:
            self.log.warning(f"Can't find callback function \"{event.tool.name}\" in {self.callbacks}")

            return ToolAnswer(
                tool_id=event.tool.tool_id,
                name=event.tool.name,
                text="function not binded"
            )

        while True:
            response: ToolAnswer = await callback(event=event)
            self.log.debug(f"Dispatching event -> {event} -> {callback} -> {response}")

            try:
                ToolResponse[response.text]

            except KeyError:
                break

            await asyncio.sleep(5)

        return response

    async def handle_tools(self, tools: list[Tool], chat_id) -> list[ToolAnswer]:
        tool_answers = [
            await self.dispatch_event(ToolEvent(tool_name=tool.name, tool=tool, chat_id=chat_id))
            for tool in tools
        ]

        return tool_answers

    async def request(self, request: SendMessageRequest) -> AgentResponse:
        self.log.debug(f"Sending request -> {request.to_dict()}")

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.endpoint}/chat",
                json=request.to_dict(),
                raise_for_status=False
            ) as resp:

                self.log.debug(f"Got response -> {await resp.text()}")

                response = await resp.json()
                agent_response = AgentResponse(**response["data"])

                return agent_response

    async def send(self, chat_id: str, text: str = None, context: dict = {}, tool_answers: list[ToolAnswer] = []) -> AgentResponse:
        agent_response = await self.request(SendMessageRequest(chat_id=chat_id, text=text, context=context, tool_answers=tool_answers))

        if agent_response.type == AgentResponseType.answer:
            return agent_response

        tool_answers = await self.handle_tools(agent_response.content, chat_id)

        return await self.send(chat_id, tool_answers=tool_answers)

    async def send_tool_states(self, tools: list[Tool], chat_id: str):
        tool_answers = await self.handle_tools(tools, chat_id)

        return await self.send(chat_id, tool_answers=tool_answers)

    async def request_tool_states(self) -> list[ChatToolState]:
        self.log.debug(f"Getting tool states")

        async with aiohttp.ClientSession() as session:
            async with session.get(
                url=f"{self.endpoint}/chat/states",
                raise_for_status=False
            ) as resp:
                self.log.debug(f"Got response -> {await resp.text()}")

                response = await resp.json()
                tool_states = list(map(lambda x: ChatToolState(**x), response))

                return tool_states

    async def handle_tool_states(self):
        tool_states = await self.request_tool_states()
        tools_by_chat = {}

        for tool_state in tool_states:
            if tool_state.chat_id not in tools_by_chat:
                tools_by_chat[tool_state.chat_id] = []

            tools_by_chat[tool_state.chat_id].append(Tool(
                id=tool_state.tool_id,
                name=tool_state.name,
                arguments=tool_state.arguments
            ))

        return await asyncio.gather(*[
            self.send_tool_states(tools_by_chat[chat_id], chat_id=chat_id)
            for chat_id in tools_by_chat
        ])




import fastapi

from .models import (
    GetStorageRequest, GetStorageResponse,
    AddStorageRequest, AddStorageResponse,
    RemoveStorageRequest, RemoveStorageResponse,
)

from AgentService.config import Config
from AgentService.types import Storage, StorageItem


storage_router = fastapi.APIRouter(prefix="/storage")


@storage_router.get("")
async def get_storage(request: GetStorageRequest) -> GetStorageResponse:
    storage = await Storage.find_one(Storage.key == request.key)
    storage_items = await StorageItem.find_many(StorageItem.storage_id == storage.id).to_list()

    return GetStorageResponse(
        data=storage_items,
        status="ok"
    )


@storage_router.post("/add")
async def add_to_storage(request: AddStorageRequest):
    key = request.key
    data = request.data

    agent = Config().agent

    if not isinstance(data, list):
        data = [data]

    storage = await Storage.find_one(Storage.key == key)
    if not storage:
        storage = await agent.create_storage(key=key)
        await storage.insert()

    for item in data:
        storage_item = StorageItem(
            storage_id=storage.id,
            data=item
        )

        existing = await StorageItem.find_one(StorageItem.key_id == storage_item.key_id)
        if existing:
            await existing.replace(item)

        else:
            await storage_item.insert()

    buffer = await Storage.get_packed(key)
    await agent.update_storage(buffer, storage.id)

    return AddStorageResponse(
        status="ok"
    )


@storage_router.post("/remove")
async def remove_from_storage(request: RemoveStorageRequest) -> RemoveStorageResponse:
    key = request.key
    data = request.data

    agent = Config().agent

    if not isinstance(data, list):
        data = [data]

    storage = await Storage.find_one(Storage.key == key)
    for item in data:
        storage_item = StorageItem(
            storage_id=storage.id,
            data=item
        )

        existing = await StorageItem.find_one(StorageItem.key_id == storage_item.key_id)
        if existing:
            await existing.delete()

        else:
            return RemoveStorageResponse(
                description=f"No such storage item -> {storage_item}",
                status="error"
            )

    storage = await Storage.find_one(Storage.key == key)
    buffer = await Storage.get_packed(key)
    await agent.update_storage(buffer, storage.id)

    return RemoveStorageResponse(
        status="ok"
    )




import fastapi
import uvicorn
from contextlib import asynccontextmanager

from AgentService.config import Config
from AgentService.types import setup_database

from .routes import (
    chat_router,
    storage_router
)


@asynccontextmanager
async def lifespan(app: fastapi.FastAPI):
    await setup_database(Config().db_uri, Config().db_name)

    yield

    return


app = fastapi.FastAPI()
app.include_router(chat_router)
app.include_router(storage_router)


def start_app():
    config = Config()
    uvicorn.run(app, host=config.app_host, port=config.app_port)




from pydantic import BaseModel


class AgentTool:
    name: str = "tool"
    description: str = "this is tool"
    parameters: BaseModel | dict = None

    @property
    def openai_model_dump(self) -> dict:
        if isinstance(self.parameters, BaseModel):
            parameters = self.parameters.model_json_schema()

        else:
            parameters = self.parameters

        if not self.parameters:
            parameters = {}

        return {
            "type": "function",
            "name": self.name,
            "description": self.description,
            "parameters": parameters
        }




from beanie import Document
from beanie.odm.documents import PydanticObjectId

from .meta import TimestampMixin

from ..enums.chat import ChatStatus, ChatToolStateStatus


class Chat(TimestampMixin, Document):
    chat_id: str
    status: ChatStatus = ChatStatus.created
    data: dict


class ChatToolState(TimestampMixin, Document):
    chat_id: PydanticObjectId
    status: ChatToolStateStatus = ChatToolStateStatus.in_progress
    tool_id: str
    name: str
    arguments: str




from beanie import Document

from .meta import TimestampMixin
from .tool import Tool, ToolAnswer

from ..enums.message import MessageType


class Message(TimestampMixin, Document):
    chat_id: str
    type: MessageType
    text: str = None
    tools: list[Tool] = None
    tool_answer: ToolAnswer = None

    @property
    def openai_model_dump(self) -> dict:
        if self.type == MessageType.tools:
            return [
                {
                    "type": "function_call",
                    "call_id": tool.tool_id,
                    "name": tool.name,
                    "arguments": tool.arguments
                }
                for tool in self.tools
            ]

        elif self.type == MessageType.tool_answer:
            return {
                "type": "function_call_output",
                "call_id": self.tool_answer.tool_id,
                "output": self.tool_answer.text
            }

        elif self.type == MessageType.assistant_skip:
            return {
                "role": "assistant",
                "content": self.text
            }

        else:
            return {
                "role": self.type.value,
                "content": self.text
            }





from beanie import Document

from .meta import TimestampMixin
from .tool import Tool

from ..enums.agent import AgentResponseType


class AgentResponse(TimestampMixin, Document):
    type: AgentResponseType
    content: str | list[Tool]




from datetime import datetime, UTC

from pydantic import BaseModel, Field
from beanie import before_event, Insert


class TimestampMixin(BaseModel):
    created_at: datetime = Field(default_factory=datetime.utcnow)

    @before_event(Insert)
    def set_created_at(self):
        self.created_at = datetime.now(UTC)
